/*
 * dspboot - simple bootloader for OMAP DSP
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>
#include <getopt.h>
#include <stdarg.h>
#include <errno.h>
#include <inttypes.h>

#define DSP_CODE_MEM		0xc3800000UL
#define DSP_CODE_SIZE		0x00700000UL

/*
 * we just need to erase shared memory header
 */
#define DSP_DATA_MEM		0xc2000000UL
#define DSP_DATA_SIZE		0x00001000UL

typedef uint8_t u8;
typedef uint32_t u32;

static int verbose;
static const char *image_file;

#if __GNUC__
# define barrier()	__asm__ __volatile__ ("" : : : "memory")
#else
# define barrier()	do { } while (0)
#endif

#if (__GNUC > 4) || (__GNUC__ == 4 && __GNUC_MINOR__ >= 1)
#define mb() __sync_synchronize()
#define rmb() mb()
#define wmb() mb()
#else
#define mb()				barrier()
#define rmb()				mb()
#define wmb()				mb()
#endif

#define __iomem

#define PSC0			0x01c10000UL
#define PSC0_SIZE		0x00001000UL
#define SYSCFG			0x01c14000UL
#define SYSCFG_SIZE		0x00001000UL

#define SYSCFG_HOST1CFG		0x0044

#define PSC0_PTCMD		0x0120
#define PSC0_PTSTAT		0x0128

#define PSC0_PDSTAT0		0x0200
#define PSC0_MDSTAT0		0x0800
#define PSC0_MDSTAT15		0x083C
#define PSC0_MDCTL15		0x0A3C

#define PSC0_PD0		1
#define PSC0_PD1		2

static void __iomem *psc0;
static void __iomem *syscfg;

#define __raw_readl(addr) (*(const volatile u32 *)(addr))
#define __raw_writel(x, addr) (*(volatile u32 *)(addr) = (x))

static inline u32 readl(const volatile void __iomem *addr)
{
	u32 ret = __raw_readl(addr);
	mb();
	return ret;
}

static inline void writel(u32 b, volatile void __iomem *addr)
{
	__raw_writel(b, addr);
	mb();
}

#define memcpy_fromio(dst, src, count) memcpy(dst, src, count)
#define memcpy_toio(dst, src, count) memcpy(dst, src, count)

static void debug(const char *fmt, ...) \
		    __attribute__((__format__ (__printf__, 1, 2)));
static void die(const char *fmt, ...) \
		    __attribute__((__format__ (__printf__, 1, 2)));

static void debug(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (verbose > 0)
		vfprintf(stderr, fmt, ap);
	va_end(ap);
}

static void die(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

static u8 image[DSP_CODE_SIZE];
static unsigned int image_len;

/*
 * xread() is the same a read(), but it automatically restarts read()
 * operations with a recoverable error (EINTR). xread()
 * DOES NOT GUARANTEE that "len" bytes is read even if the data is available.
 */
static ssize_t xread(int fd, void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = read(fd, buf, len);
		if ((nr < 0) && (errno == EINTR))
			continue;
		return nr;
	}
}

static void load_file_or_die(const char *fname)
{
	int fd;
	ssize_t ret;

	fd = open(fname, O_RDONLY);
	if (fd == -1) {
		perror("open");
		exit(EXIT_FAILURE);
	}

	do {
		ret = xread(fd, image + image_len, DSP_CODE_SIZE - image_len);
		if (ret < 0) {
			perror("read");
			exit(EXIT_FAILURE);
		}
		image_len += ret;
	} while (ret > 0);

	debug("image size: %d bytes\n", image_len);
	if (image_len >= DSP_CODE_SIZE)
		die("error: image too big\n");
	close(fd);
}

static void *iomap_or_die(unsigned long base, unsigned long len)
{
	unsigned long page_size;
	unsigned long offset;
	char *p;
	int fd;

	page_size = sysconf(_SC_PAGE_SIZE);
	debug("map %lx len: %ld\n", base, len);
	offset = base & (page_size - 1);
	debug("offset %ld\n", offset);
	len = (len + offset + page_size - 1) & ~(page_size - 1);

	fd = open("/dev/mem", O_RDWR);
	if (fd == -1) {
		perror("error: cannot open /dev/mem");
		exit(EXIT_FAILURE);
	}
	debug("mapping %ld bytes at %08lx\n", len, base - offset);
	p = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd,
			base - offset);
	if (p == MAP_FAILED) {
		perror("error: cannot map DSP code memory");
		exit(EXIT_FAILURE);
	}
	close(fd);
	return p + offset;
}

static void iounmap(void *p)
{
}

/*
 * we always write all DSP code memory
 */
static void copy_to_dsp_mem(const void *image, unsigned int len)
{
	void *p = iomap_or_die(DSP_CODE_MEM, DSP_CODE_SIZE);

	memcpy_toio(p, image, DSP_CODE_SIZE);
	mb();
	iounmap(p);
}

static void clear_data_mem(void)
{
	void *p = iomap_or_die(DSP_DATA_MEM, DSP_DATA_SIZE);

	memset(p, 0, DSP_DATA_SIZE);
	mb();
	iounmap(p);
}

static void dsp_on(void)
{
	u32 x;

	debug("waiting on PTSTAT...\n");
	do {
		x = readl(psc0 + PSC0_PTSTAT);
	} while (x & PSC0_PD1);
	debug("OK\n");

	writel(3, psc0 + PSC0_MDCTL15);
	writel(PSC0_PD1, psc0 + PSC0_PTCMD);

	debug("waiting on PTSTAT...\n");
	do {
		x = readl(psc0 + PSC0_PTSTAT);
	} while (x & PSC0_PD1);
	debug("OK\n");

	debug("waiting on MDSTAT15...\n");
	do {
		x = readl(psc0 + PSC0_MDSTAT15);
		fprintf(stderr, "x = %08x\n", x);
		usleep(100000);
	} while ((x & 0x3f) != 3);
	debug("OK\n");
}

static void psc_status(void)
{
	unsigned int i;
	u32 x;

	for (i = 0; i < 2; i++) {
		x = readl(psc0 + PSC0_PDSTAT0 + 4 * i);
		fprintf(stderr, "PSC0_PDSTAT%d = %08x\n", i, x);
	}

	for (i = 0; i < 16; i++) {
		x = readl(psc0 + PSC0_MDSTAT0 + 4 * i);
		fprintf(stderr, "PSC0_MDSTAT%d = %08x\n", i, x);
	}
}

static void dsp_stop(void)
{
	u32 x;

	x = readl(psc0 + PSC0_MDCTL15);
	x &= ~0x100;
	writel(x, psc0 + PSC0_MDCTL15);
	x = readl(psc0 + PSC0_MDCTL15);
	debug("PSC0_MDCTL15 = %08x\n", x);
}

static void boot_dsp(void)
{
	u32 x;

	x = readl(syscfg + SYSCFG_HOST1CFG);
	debug("SYSCFG_HOST1CFG = %08x\n", x);
	writel(DSP_CODE_MEM, syscfg + SYSCFG_HOST1CFG);
	x = readl(syscfg + SYSCFG_HOST1CFG);
	debug("SYSCFG_HOST1CFG = %08x\n", x);

	dsp_on();

	x = readl(psc0 + PSC0_MDCTL15);
	x |= 0x100;
	writel(x, psc0 + PSC0_MDCTL15);
	x = readl(psc0 + PSC0_MDCTL15);
	debug("PSC0_MDCTL15 = %08x\n", x);
}

static const char copyright_string[] =
"dspboot " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [--] <FILE>\n"
"Options:\n"
"  -V, --version              show version\n"
"  -h, --help                 show this help\n"
"  -v, --verbose              be verbose\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "dspboot" : name);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
}

static void parse_command_line(int argc, char **argv)
{
	static struct option long_options[] = {
		{"help", 0, NULL, 'h'},
		{"verbose", 0, NULL, 'v'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, argv, "hvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'h':
			show_usage(argv[0]);
			exit(EXIT_SUCCESS);
		case 'v':
			verbose++;
			break;
		case 'V':
			show_version();
			exit(EXIT_SUCCESS);
		default:
			die("Unknown option: %d.\n", c);
		}
	}

	if (optind < argc)
		image_file = argv[optind++];
}

int main(int argc, char **argv)
{
	if (argc <= 1) {
		show_usage(argv[0]);
		exit(EXIT_FAILURE);
	}
	parse_command_line(argc, argv);
	if (!image_file || !*image_file)
		die("error: FILE argument is required\n");

	syscfg = iomap_or_die(SYSCFG, SYSCFG_SIZE);
	psc0 = iomap_or_die(PSC0, PSC0_SIZE);

	if (verbose)
		psc_status();
	dsp_stop();
	load_file_or_die(image_file);
	copy_to_dsp_mem(image, image_len);
	clear_data_mem();

	boot_dsp();
	exit(EXIT_SUCCESS);
}
