/*
 * dspconfig - update configuration for DSP
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>
#include <getopt.h>
#include <stdarg.h>
#include <errno.h>
#include <inttypes.h>

#define DSP_CONFIG_MEM		0xc3f00000UL
#define DSP_CONFIG_SIZE		0x00100000UL
#define DSP_FOOTER_SIZE		0x00001000UL

#define DSP_CONFIG_MAGIC	0x000
#define DSP_CONFIG_MAGIC_VALUE	0x43487550
#define DSP_CONFIG_SEQLOCK	0x004
#define DSP_CONFIG_VERSION	0x008
#define DSP_CONFIG_LEN		0x00c

typedef uint8_t u8;
typedef uint32_t u32;

static int verbose;
static const char *config_file;

#if __GNUC__
# define barrier()	__asm__ __volatile__ ("" : : : "memory")
#else
# define barrier()	do { } while (0)
#endif

#if (__GNUC > 4) || (__GNUC__ == 4 && __GNUC_MINOR__ >= 1)
#define mb() __sync_synchronize()
#define rmb() mb()
#define wmb() mb()
#else
#define mb()				barrier()
#define rmb()				mb()
#define wmb()				mb()
#endif

#define __iomem

#define __raw_readl(addr) (*(const volatile u32 *)(addr))
#define __raw_writel(x, addr) (*(volatile u32 *)(addr) = (x))

static inline u32 readl(const volatile void __iomem *addr)
{
	u32 ret = __raw_readl(addr);
	mb();
	return ret;
}

static inline void writel(u32 b, volatile void __iomem *addr)
{
	__raw_writel(b, addr);
	mb();
}

#define memcpy_fromio(dst, src, count) memcpy(dst, src, count)
#define memcpy_toio(dst, src, count) memcpy(dst, src, count)

static u8 __iomem *config_mem;
static u8 __iomem *config_footer;

static void debug(const char *fmt, ...) \
		    __attribute__((__format__ (__printf__, 1, 2)));
static void die(const char *fmt, ...) \
		    __attribute__((__format__ (__printf__, 1, 2)));

static void debug(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (verbose > 0)
		vfprintf(stderr, fmt, ap);
	va_end(ap);
}

static void die(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

static u8 config[DSP_CONFIG_SIZE - DSP_FOOTER_SIZE];
static unsigned int config_len;

/*
 * xread() is the same a read(), but it automatically restarts read()
 * operations with a recoverable error (EINTR). xread()
 * DOES NOT GUARANTEE that "len" bytes is read even if the data is available.
 */
static ssize_t xread(int fd, void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = read(fd, buf, len);
		if ((nr < 0) && (errno == EINTR))
			continue;
		return nr;
	}
}

static void load_file_or_die(const char *fname)
{
	int fd;
	ssize_t ret;

	fd = open(fname, O_RDONLY);
	if (fd == -1) {
		perror("open");
		exit(EXIT_FAILURE);
	}

	do {
		ret = xread(fd, config + config_len,
				DSP_CONFIG_SIZE - config_len);
		if (ret < 0) {
			perror("read");
			exit(EXIT_FAILURE);
		}
		config_len += ret;
	} while (ret > 0);

	debug("config size: %d bytes\n", config_len);
	if (config_len >= DSP_CONFIG_SIZE - DSP_FOOTER_SIZE)
		die("error: config too big\n");
	close(fd);
}

static void *iomap_or_die(unsigned long base, unsigned long len)
{
	unsigned long page_size;
	unsigned long offset;
	char *p;
	int fd;

	page_size = sysconf(_SC_PAGE_SIZE);
	debug("map %lx len: %ld\n", base, len);
	offset = base & (page_size - 1);
	debug("offset %ld\n", offset);
	len = (len + offset + page_size - 1) & ~(page_size - 1);

	fd = open("/dev/mem", O_RDWR);
	if (fd == -1) {
		perror("error: cannot open /dev/mem");
		exit(EXIT_FAILURE);
	}
	debug("mapping %ld bytes at %08lx\n", len, base - offset);
	p = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd,
			base - offset);
	if (p == MAP_FAILED) {
		perror("error: cannot map DSP code memory");
		exit(EXIT_FAILURE);
	}
	close(fd);
	return p + offset;
}

static void iounmap(void *p)
{
}

static const char copyright_string[] =
"dspconfig " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [--] <FILE>\n"
"Options:\n"
"  -V, --version              show version\n"
"  -h, --help                 show this help\n"
"  -v, --verbose              be verbose\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "dspboot" : name);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
}

static void parse_command_line(int argc, char **argv)
{
	static struct option long_options[] = {
		{"help", 0, NULL, 'h'},
		{"verbose", 0, NULL, 'v'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, argv, "hvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'h':
			show_usage(argv[0]);
			exit(EXIT_SUCCESS);
		case 'v':
			verbose++;
			break;
		case 'V':
			show_version();
			exit(EXIT_SUCCESS);
		default:
			die("Unknown option: %d.\n", c);
		}
	}

	if (optind < argc)
		config_file = argv[optind++];
}

/*
 * we always write all DSP code memory
 */
static void copy_to_dsp_mem(const void *config, unsigned int len)
{
	memcpy_toio(config_mem, config, DSP_CONFIG_SIZE - DSP_FOOTER_SIZE);
	mb();
}

static void config_init(void)
{
	writel(0, config_footer + DSP_CONFIG_VERSION);
	writel(config_len, config_footer + DSP_CONFIG_LEN);
	writel(DSP_CONFIG_MAGIC_VALUE, config_footer + DSP_CONFIG_MAGIC);
}

int main(int argc, char **argv)
{
	u32 seqlock;

	if (argc <= 1) {
		show_usage(argv[0]);
		exit(EXIT_FAILURE);
	}
	parse_command_line(argc, argv);
	if (!config_file || !*config_file)
		die("error: FILE argument is required\n");

	load_file_or_die(config_file);

	config_mem = iomap_or_die(DSP_CONFIG_MEM, DSP_CONFIG_SIZE);
	config_footer = config_mem + DSP_CONFIG_SIZE - DSP_FOOTER_SIZE;

	debug("config mapped at %p, footer %p\n", config_mem, config_footer);

	/* lock seqlock */
	seqlock = readl(config_footer + DSP_CONFIG_SEQLOCK);
	seqlock |= 1;
	debug("lock: seqlock %ld\n", (unsigned long) seqlock);
	writel(seqlock, config_footer + DSP_CONFIG_SEQLOCK);

	debug("copying config_len bytes\n");
	copy_to_dsp_mem(config, config_len);
	config_init();

	/* unlock seqlock */
	seqlock++;
	debug("unlock: seqlock %ld\n", (unsigned long) seqlock);
	writel(seqlock, config_footer + DSP_CONFIG_SEQLOCK);
	
	iounmap(config_mem);

	exit(EXIT_SUCCESS);
}
